from collections.abc import Generator, Iterable, Sequence
from contextlib import contextmanager
from functools import partial
from typing import Any, Literal

from bson import ObjectId
from typing_extensions import Self

from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsIngester import AbsIngester
from ampel.content.DataPoint import DataPoint
from ampel.content.JournalRecord import JournalRecord
from ampel.content.StockDocument import StockDocument
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.enum.JournalActionCode import JournalActionCode
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.MongoStockUpdater import BaseStockUpdater
from ampel.protocol.StockIngesterProtocol import StockIngesterProtocol
from ampel.queue.AbsProducer import AbsProducer
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.types import ChannelId, StockId, Tag

# ruff: noqa: SLF001

class QueueIngester(AbsIngester):
    producer: UnitModel

    class QueueStockUpdater(BaseStockUpdater):
        def __init__(
            self,
            queue: "QueueIngester",
            tier: Literal[-1, 0, 1, 2, 3],
            run_id: int,
            process_name: str,
            extra_tag: None | Tag | Sequence[Tag] = None,
            bump_updated: bool = True,

        ) -> None:
            super().__init__(
                tier=tier, run_id=run_id, process_name=process_name, extra_tag=extra_tag
            )
            self.queue = queue
            self.bump_updated = bump_updated


        def add_journal_record(
            self,
            stock: StockId | Sequence[StockId],
            jattrs: None | JournalAttributes = None,
            tag: None | Tag | Sequence[Tag] = None,
            name: None | str | Sequence[str] = None,
            trace_id: None | dict[str, int] = None,
            action_code: None | JournalActionCode = None,
            doc_id: None | ObjectId = None,
            unit: None | int | str = None,
            channel: None | ChannelId | Sequence[ChannelId] = None,
            now: None | int | float = None,
        ) -> JournalRecord:
            """
            Add a journal record to the stock document(s) identified by the input stock id(s)
            """
            jrec = self.new_journal_record(
                unit=unit,
                channels=channel,
                action_code=action_code,
                doc_id=doc_id,
                jattrs=jattrs,
                trace_id=trace_id,
                now=now,
            )

            names = {name} if isinstance(name, str) else set(name) if name else None
            tags = {tag} if isinstance(tag, Tag) else set(tag) if tag else None
            channels = [channel] if isinstance(channel, ChannelId) else channel if channel else []

            for doc in self._match_stocks(
                {stock} if isinstance(stock, StockId) else set(stock)
            ):
                doc["journal"] = [*doc.get("journal", []), jrec]
                if names:
                    doc["name"] = list(names.union(doc.get("name", [])))
                if tags:
                    doc["tag"] = list(tags.union(doc.get("tag", [])))
                
                if self.bump_updated:
                    for chan in ("any", *channels):
                        self._update_ts(doc, chan, jrec["ts"])

            return jrec

        @staticmethod
        def _update_ts(doc: StockDocument, channel: ChannelId, ts: int | float) -> None:
            if channel in doc["ts"]:
                if "upd" in doc["ts"][channel]:
                    doc["ts"][channel]["upd"] = max(ts, doc["ts"][channel]["upd"])
                else:
                    doc["ts"][channel]["upd"] = ts
            else:
                doc["ts"][channel] = {"upd": ts}

        def add_name(self, stock: StockId, name: str | Sequence[str]) -> None:
            names = {name} if isinstance(name, str) else set(name)
            for doc in self._match_stocks({stock}):
                doc["name"] = list(names.union(doc.get("name", [])))

        def add_tag(
            self, stock: StockId | Sequence[StockId], tag: Tag | Sequence[Tag]
        ) -> None:
            tags = {tag} if isinstance(tag, Tag) else set(tag)
            for doc in self._match_stocks(
                {stock} if isinstance(stock, StockId) else set(stock)
            ):
                doc["tag"] = list(tags.union(doc.get("tag", [])))

        def _match_stocks(
            self, stock_ids: set[StockId]
        ) -> Generator[StockDocument, None, None]:
            for stock in self.queue._item.stock:
                if stock["stock"] in stock_ids:
                    yield stock

    class QueueStockIngester(AbsDocIngester[StockDocument]):
        queue: "QueueIngester"
        update: "QueueIngester.QueueStockUpdater"

        def ingest(self, doc: StockDocument) -> None:
            self.queue._add_stock(doc)

    class QueueT0Ingester(AbsDocIngester[DataPoint]):
        queue: "QueueIngester"

        def ingest(self, doc: DataPoint) -> None:
            self.queue._add_t0(doc)

    class QueueT1Ingester(AbsDocIngester[T1Document]):
        queue: "QueueIngester"

        def ingest(self, doc: T1Document) -> None:
            self.queue._add_t1(doc)

    class QueueT2Ingester(AbsDocIngester[T2Document]):
        queue: "QueueIngester"

        def ingest(self, doc: T2Document) -> None:
            self.queue._add_t2(doc)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._producer = self.context.loader.new(self.producer, unit_type=AbsProducer)

        self._stock = self.QueueStockIngester(
            queue=self,
            update=self.QueueStockUpdater(
                queue=self,
                tier=self.tier,
                run_id=self.run_id,
                process_name=self.process_name,
            ),
        )
        self._t0 = self.QueueT0Ingester(queue=self)
        self._t1 = self.QueueT1Ingester(queue=self)
        self._t2 = self.QueueT2Ingester(queue=self)

        self._item = AbsProducer.Item.new()

    def __enter__(self) -> "Self":
        self._producer.__enter__()
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback) -> bool | None:
        return self._producer.__exit__(exc_type, exc_value, traceback)

    @contextmanager
    def group(self, acknowledge_messages: None | Iterable[Any] = None) -> Generator:
        """
        Ensure that updates issued in this context are grouped together
        """
        yield
        item = self._swap_buffer()
        if item or acknowledge_messages:
            self._producer.produce(
                item,
                partial(self.acknowledge_callback, acknowledge_messages)
                if self.acknowledge_callback and acknowledge_messages is not None
                else None,
            )

    def _swap_buffer(self) -> AbsProducer.Item:
        prev = self._item
        self._item = AbsProducer.Item.new()
        return prev

    def _add_stock(self, doc: StockDocument) -> None:
        self._item.stock.append(doc)

    def _add_t0(self, doc: DataPoint) -> None:
        self._item.t0.append(doc)

    def _add_t1(self, doc: T1Document) -> None:
        self._item.t1.append(doc)

    def _add_t2(self, doc: T2Document) -> None:
        self._item.t2.append(doc)

    @property
    def stock(self) -> StockIngesterProtocol:
        return self._stock

    @property
    def t0(self) -> AbsDocIngester[DataPoint]:
        return self._t0

    @property
    def t1(self) -> AbsDocIngester[T1Document]:
        return self._t1

    @property
    def t2(self) -> AbsDocIngester[T2Document]:
        return self._t2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/test/test_T3Processor.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  25.07.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator
from typing import Any

import pytest

from ampel.abstract.AbsT3Unit import AbsT3Unit, T3Send
from ampel.content.StockDocument import StockDocument
from ampel.content.T2Document import T2Document
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.enum.DocumentCode import DocumentCode
from ampel.enum.EventCode import EventCode
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.struct.StockAttributes import StockAttributes
from ampel.struct.T3Store import T3Store
from ampel.t3.T3Processor import T3Processor
from ampel.test.dummy import DummyStateT2Unit
from ampel.util.config import get_unit_confid
from ampel.view.SnapView import SnapView


class Mutineer(AbsT3Unit):
    raise_on_process: bool = False

    def process(self, views, t3s=None):
        if self.raise_on_process:
            raise ValueError


def mutineer_process(config=None):

    return {
        "supply": {
            "unit": "T3DefaultBufferSupplier",
            "config": {
                "select": {"unit": "T3StockSelector"},
                "load": {
                    "unit": "T3SimpleDataLoader",
                    "config": {
                        "directives": [{"col": "stock"}]
                    }
                }
            }
        },
        "stage": {
            "unit": "T3SimpleStager",
            "config": {
                "execute": [{"unit": "Mutineer", "config": config or {}}]
            }
        }
    }


@pytest.mark.usefixtures("_ingest_stock")
@pytest.mark.parametrize(
    ("config","expect_success"),
    [
        ({}, True),
        ({"raise_on_process": True}, False),
    ]
)
def test_unit_raises_error(
    dev_context: DevAmpelContext, config, expect_success
):
    """Run is marked failed if units raise an exception"""
    dev_context.register_unit(Mutineer)
    t3 = T3Processor(context=dev_context, process_name="test", raise_exc=False, **mutineer_process(config))
    t3.run()
    assert dev_context.db.get_collection('event').count_documents({}) == 1
    event = dev_context.db.get_collection('event').find_one({})
    assert event
    assert event["run"] == 1
    assert event["code"] == EventCode.OK.value if expect_success else EventCode.EXCEPTION


@pytest.mark.usefixtures("_ingest_stock")
def test_view_generator(integration_context: DevAmpelContext):

    class SendySend(AbsT3Unit):
        raise_on_process: bool = False

        def process(self, gen: Generator[SnapView, T3Send, None], t3s: None | T3Store = None):
            for view in gen:
                gen.send(
                    (
                        view.id,
                        StockAttributes(
                            tag="TAGGYTAG",
                            name="floopsy",
                            journal=JournalAttributes(extra={"foo": "bar"}),
                        ),
                    )
                )

    integration_context.register_unit(SendySend)

    t3 = T3Processor(
        context=integration_context,
        raise_exc=True,
        process_name="t3",
        supply = {
            "unit": "T3DefaultBufferSupplier",
            "config": {
                "select": {"unit": "T3StockSelector"},
                "load": {
                    "unit": "T3SimpleDataLoader",
                    "config": {
                        "directives": [{"col": "stock"}]
                    }
                }
            }
        },
        stage = {
            "unit": "T3SimpleStager",
            "config": {
                "execute": [{"unit": "SendySend"}]
            }
        }
    )
    t3.run()

    stock = integration_context.db.get_collection("stock").find_one()
    assert stock
    assert "TAGGYTAG" in stock["tag"]
    assert "floopsy" in stock["name"]
    assert len(entries := [jentry for jentry in stock["journal"] if jentry["tier"] == 3]) == 1
    jentry = entries[0]
    assert jentry["extra"] == {"foo": "bar"}

@pytest.mark.usefixtures("_ingest_stock")
def test_empty_generator(integration_context: DevAmpelContext):
    """
    Empty selection returns cleanly, rather than raising
    """
    t3 = T3Processor(
        context=integration_context,
        raise_exc=True,
        process_name="t3",
        supply = {
            "unit": "T3DefaultBufferSupplier",
            "config": {
                "select": {
                    "unit": "T3StockSelector",
                    # ensure that no stocks will be selected
                    "config": {"channel": "nonesuch"}
                },
                "load": {
                    "unit": "T3SimpleDataLoader",
                    "config": {
                        "directives": [{"col": "stock"}],
                    }
                }
            }
        },
        stage = {
            "unit": "T3SimpleStager",
            "config": {
                "execute": [{"unit": "DemoT3Unit"}]
            }
        }
    )
    t3.run()


class ViewExaminer(AbsT3Unit):
    class DidAThing(Exception):
        ...

    expected_config: dict[str, Any]

    def process(self, views, t3s=None):
        for view in views:
            for t2 in view.t2:
                assert t2.config == self.expected_config
                raise self.DidAThing


@pytest.mark.parametrize("num_units", [1, 2])
def test_t2_config_resolution(mock_context: DevAmpelContext, num_units: int):
    """ T3 stagers pass through resolved config ids """

    for unit in (ViewExaminer, DummyStateT2Unit):
        mock_context.register_unit(unit)

    expected_config = {"foo": 42}
    stock: StockDocument = {"stock": 0}
    confid = get_unit_confid(mock_context.loader, "DummyStateT2Unit", expected_config)
    mock_context.add_conf_id(confid, expected_config)
    t2: T2Document = {
        "unit": "DummyStateT2Unit",
        "config": confid,
        "stock": 0,
        "link": 0,
        "channel": "TEST",
        "meta": [],
        "body": [],
        "code": DocumentCode.OK,
    }
    mock_context.db.get_collection("stock").insert_one(dict(stock))
    mock_context.db.get_collection("t2").insert_one(dict(t2))

    t3 = T3Processor(
        context=mock_context,
        raise_exc=True,
        process_name="t3",
        supply = {
            "unit": "T3DefaultBufferSupplier",
            "config": {
                "select": {
                    "unit": "T3StockSelector",
                },
                "load": {
                    "unit": "T3SimpleDataLoader",
                    "config": {
                        "directives": [{"col": "t2"}],
                    }
                }
            }
        },
        stage = {
            "unit": "T3SimpleStager",
            "config": {
                "execute": [
                    {
                        "unit": "ViewExaminer",
                        "config": {"expected_config": expected_config},
                    }
                ]
                * num_units
            }
        }
    )
    with pytest.raises(ViewExaminer.DidAThing):
        t3.run()




from types import TracebackType

from typing_extensions import Self

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod


class AbsContextManager(AmpelABC, abstract=True):
    """
    AmpelABC version of contextlib.AbstractContextManager
    """

    def __enter__(self) -> "Self":
        return self

    @abstractmethod
    def __exit__(self, exc_type: type[BaseException], exc_val: BaseException, exc_tb: TracebackType) -> None | bool:
        ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t4/T4Processor.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                02.04.2023
# Last Modified Date:  04.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from datetime import datetime, timezone
from typing import Annotated

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsT4ControlUnit import AbsT4ControlUnit
from ampel.abstract.AbsT4Unit import AbsT4Unit
from ampel.content.T4Document import T4Document
from ampel.core.DocBuilder import DocBuilder
from ampel.core.EventHandler import EventHandler
from ampel.log import SHOUT, AmpelLogger, LogFlag
from ampel.model.UnitModel import UnitModel
from ampel.types import ChannelId


class T4Processor(AbsEventUnit, DocBuilder):

	execute: Annotated[Sequence[UnitModel], AbsT4Unit | AbsT4ControlUnit]
	channel: None | ChannelId = None

	def __init__(self, **kwargs) -> None:
		if isinstance(kwargs.get('execute', []), dict):
			kwargs['execute'] = [kwargs['execute']]
		super().__init__(**kwargs)


	def proceed(self, event_hdlr: EventHandler) -> None:

		event_hdlr.set_tier(4)
		logger = AmpelLogger.from_profile(
			self.context, self.log_profile, event_hdlr.get_run_id(),
			base_flag = LogFlag.T4 | LogFlag.CORE | self.base_log_flag,
			force_refresh = True
		)

		try:

			for um in self.execute:

				t4_unit_info = self.context.config.get(f'unit.{um.unit}', dict, raise_exc=True)
				if 'AbsT4Unit' in t4_unit_info['base']:
					t4_unit: AbsT4Unit | AbsT4ControlUnit = self.context.loader.new_safe_logical_unit(
						um=um, unit_type=AbsT4Unit, logger=logger,
						_chan=self.channel # unclear if multiple chans are supported
					)

				else:
					t4_unit = self.context.loader.new_context_unit(
						model = um,
						context = self.context,
						sub_type = AbsT4ControlUnit,
						logger = logger,
						event_hdlr = event_hdlr,
						channel = self.channel
					)

				logger.log(SHOUT, f'Running {self.process_name}', extra={'unit': t4_unit.__class__.__name__})

				ts = datetime.now(tz=timezone.utc).timestamp()

				if isinstance(t4_unit, AbsT4Unit):

					ret1 = t4_unit.do()
					if (buf_hdlr := getattr(t4_unit, '_buf_hdlr', None)) and buf_hdlr.buffer:
						buf_hdlr.forward(logger)

					if not ret1:
						continue

					t4ds: Sequence[T4Document] = [
						self.craft_doc(self.context, event_hdlr, t4_unit, ret1, ts, doc_type=T4Document)
					]

				elif isinstance(t4_unit, AbsT4ControlUnit):
					t4ds = list(t4_unit.do())

				# T4Document to be included in t4 collection
				for t4d in t4ds:

					if 'meta' not in t4d:
						raise ValueError("Invalid T4Document")

					t4d['meta']['traceid'] = {'t4processor': self._trace_id}
					if event_hdlr.job_sig:
						t4d['meta']['jobid'] = event_hdlr.job_sig

					self.context.db.get_collection('t4').insert_one(t4d)

		except Exception as e:
			event_hdlr.handle_error(e, logger)

		# Feedback
		logger.log(SHOUT, f'Done running {self.process_name}')
		logger.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ingest/IngestBody.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                05.06.2020
# Last Modified Date:  27.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.ingest.MuxModel import MuxModel
from ampel.model.ingest.T1Combine import T1Combine
from ampel.model.ingest.T1CombineCompute import T1CombineCompute
from ampel.model.ingest.T1CombineComputeNow import T1CombineComputeNow
from ampel.model.ingest.T2Compute import T2Compute


class IngestBody(AmpelBaseModel):
	"""
    stock_t2
    point_t2        <- based on input dps list
    combine         <- same
      state_t2      <- based on dps list returned by combine
      point_t2      <- same
    mux             <- based on input dps list
      insert
        point_t2    <- based on <datapoints insert> result from muxer unit (first list)
      combine
        state_t2    <- based on <datapoints combine> result from muxer unit (second list)
        point_t2    <- same
	"""

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`stocks <ampel.content.StockDocument.StockDocument>`
	stock_t2: None | Sequence[T2Compute] = None

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`datapoints <ampel.content.DataPoint.DataPoint>`
	#: based on alert content
	point_t2: None | Sequence[T2Compute] = None

	#: Create :class:`compounds <ampel.content.T1Document.T1Document>` from
	#: combined :class:`datapoints <ampel.content.DataPoint.DataPoint>` and the
	#: associated :class:`T2 documents <ampel.content.T2Document.T2Document>`
	combine: None | Sequence[T1Combine | T1CombineCompute | T1CombineComputeNow] = None

	#: Include additional material (such as datapoints from the DB)
	mux: None | MuxModel = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/alter/HashT2Config.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                05.04.2023
# Last Modified Date:  05.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.abstract.AbsConfigUpdater import AbsConfigUpdater
from ampel.config.builder.ProcessMorpher import ProcessMorpher
from ampel.config.collector.T02ConfigCollector import T02ConfigCollector
from ampel.core.AmpelContext import AmpelContext
from ampel.log.AmpelLogger import AmpelLogger


class HashT2Config(AbsConfigUpdater):
	""" Hashes t2 configs recursively """

	def alter(self, context: AmpelContext, content: dict[str, Any], logger: AmpelLogger) -> dict[str, Any]:

		pm = ProcessMorpher(process={}, logger=logger, verbose=bool(logger.verbose), proc_name='<off-config>')
		cc = T02ConfigCollector(conf_section="confid")
		# Note: *process* templating could be possible there as well, add if ever needed
		ac = context.config._config  # noqa: SLF001
		pm.hash_t2_config(
			{'unit': ac['unit'], 'alias': ac['alias'], 'confid': cc},
			target=content['directives'] if 'directives' in content else content['config']['directives']
		)

		s = []
		for conf_id, conf in cc.items():
			if conf_id not in context.db.conf_ids:
				s.append(conf_id)
				context.db.add_conf_id(conf_id, conf)
				dict.__setitem__(ac["confid"], conf_id, conf)

		if logger.verbose:
			logger.info("DB updated with new conf ids", extra={'confids': s})

		return content



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/AbsMongoIngester.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.03.2020
# Last Modified Date:  09.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer


class HasUpdatesBuffer(AmpelBaseModel):

	updates_buffer: DBUpdatesBuffer



from collections.abc import Iterable
from contextlib import contextmanager
from time import time
from typing import Any

import pytest
from pymongo.errors import OperationFailure
from pytest_mock import MockerFixture

from ampel.content.T2Document import T2Document
from ampel.core.AmpelContext import AmpelContext
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.enum.DocumentCode import DocumentCode
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
from ampel.model.UnitModel import UnitModel
from ampel.queue.QueueIngester import AbsProducer, QueueIngester
from ampel.t2.T2QueueWorker import AbsConsumer, QueueItem, T2QueueWorker
from ampel.t2.T2Worker import T2Worker
from ampel.test.conftest import make_tied_ingestion_handler
from ampel.test.dummy import DummyPointT2Unit


@contextmanager
def collect_diff(store):
    store.clear()
    before = {}
    for metric in AmpelMetricsRegistry.registry().collect():
        for sample in metric.samples:
            key = (sample.name, tuple(sample.labels.items()))
            before[key] = sample.value

    delta = {}
    yield
    for metric in AmpelMetricsRegistry.registry().collect():
        for sample in metric.samples:
            key = (sample.name, tuple(sample.labels.items()))
            delta[key] = sample.value - (before.get(key) or 0)
    store.update(delta)


@pytest.mark.usefixtures("_ingest_stock_t2")
def test_metrics(integration_context):
    t2 = T2Worker(context=integration_context, raise_exc=True, process_name="t2")

    stats = {}
    with collect_diff(stats):
        assert t2.run() == 1

    assert (
        stats[("ampel_t2_docs_processed_total", (("unit", "DummyStockT2Unit"), ('code', 'OK')))] == 1
    )
    assert stats[("ampel_t2_latency_seconds_sum", (("unit", "DummyStockT2Unit"),))] > 0


@pytest.mark.parametrize("config", [None, {"raise_exc": True}])
def test_error_reporting(integration_context: DevAmpelContext, config):
    integration_context.register_unit(DummyPointT2Unit)
    # DummyPointT2Unit will raise an error on the malformed T0 doc
    integration_context.db.get_collection("t0").insert_one({"id": 42, "stock": 42})
    channels = ["channel_a", "channel_b"]
    doc: T2Document = {
        "unit": "DummyPointT2Unit",
        "code": DocumentCode.NEW,
        "config": config,
        "col": "t0",
        "stock": 42,
        "link": 42,
        "channel": channels,
        "meta": [
            {"ts": 0, "tier": 2}
        ],
        "body": []
    }
    integration_context.db.get_collection("t2").insert_one(doc)
    t2 = T2Worker(context=integration_context, raise_exc=False, process_name="t2", run_dependent_t2s=True)
    assert t2.run() == 1
    assert (doc := integration_context.db.get_collection("t2").find_one({})) # type: ignore[assignment]
    assert doc["code"] == DocumentCode.EXCEPTION
    assert (trouble := integration_context.db.get_collection('trouble').find_one({}))
    if config is None:
        assert trouble["channel"] == channels
        assert trouble.get("msg") is None
    else:
        assert trouble["extra"]["msg"] == "Could not instantiate unit"


def test_tied_t2s(integration_context, ingest_tied_t2):
    assert (num_dps := integration_context.db.get_collection("t0").count_documents({}))
    t2 = T2Worker(
        context=integration_context, raise_exc=True, process_name="t2", run_dependent_t2s=True
    )

    num_docs = t2.run()
    t2 = integration_context.db.get_collection("t2")
    assert t2.count_documents({}) == num_docs

    if "point" in ingest_tied_t2.param.lower():
        assert num_docs == 1 + num_dps
        assert (
            next(t2.find({"unit": "DummyPointT2Unit"}).sort([("link", -1)]).limit(1))[
                "body"
            ][0]["thing"]
            == 3
        )
        assert (
            next(
                t2.find({"unit": "DummyTiedStateT2Unit"}).sort([("link", -1)]).limit(1)
            )["body"][0]["thing"]
            == 2 * 3
        )
    elif "stock" in ingest_tied_t2.param.lower():
        assert num_docs == 2
        assert (
            t2.find_one({"unit": "DummyStockT2Unit"})["body"][0]["id"] == "stockystock"
        )
        assert (
            t2.find_one({"unit": "DummyTiedStateT2Unit"})["body"][0]["id"]
            == 2 * "stockystock"
        )
    else:
        assert num_docs == 2
        assert t2.find_one({"unit": "DummyStateT2Unit"})["body"][0]["len"] == num_dps
        assert (
            t2.find_one({"unit": "DummyTiedStateT2Unit"})["body"][0]["len"]
            == 2 * num_dps
        )


def test_slow_dependency(
    integration_context: AmpelContext, ingest_tied_t2, mocker: MockerFixture
):
    """
    Simulate a race conditions between parallel T2 workers
    """
    assert integration_context.db.get_collection("t0").count_documents({})
    t2 = T2Worker(
        context=integration_context,
        raise_exc=True,
        process_name="t2",
        run_dependent_t2s=True,
        backoff_on_retry=[{"jitter": False, "factor": 10}],
    )

    # num_docs = t2.run()
    col = integration_context.db.get_collection("t2")

    # set upstream docs to RUNNING, simulating the effect of
    # a parallel worker picking them up
    assert (
        col.update_many(
            {"unit": {"$not": {"$regex": ".*Tied.*"}}},
            {"$set": {"code": DocumentCode.RUNNING}},
        ).modified_count
        > 0
    )

    # should only find depdendent doc once.
    try:
        assert t2.run() == 1
    except OperationFailure as exc:
        if str(exc) == "Unrecognized expression '$last'":
            # see:
            # - https://github.com/mongomock/mongomock/pull/734
            # - https://github.com/mongomock/mongomock/pull/770
            pytest.xfail("mongomock doesn't support $last (yet)")
        else:
            raise
    assert (
        col.count_documents({"code": DocumentCode.T2_PENDING_DEPENDENCY}) == 1
    ), "exactly 1 dependent doc marked as pending"
    dependent_doc = col.find_one({"code": DocumentCode.T2_PENDING_DEPENDENCY})
    assert dependent_doc is not None
    meta = dependent_doc["meta"][-1]
    assert t2.backoff_on_retry is not None
    assert meta["retry_after"] == meta["ts"] + t2.backoff_on_retry[0].factor
    assert t2.run() == 0, "no more docs to run"

    # run upstream docs
    while (
        doc := col.find_one_and_update(
            {"code": DocumentCode.RUNNING}, {"$set": {"code": DocumentCode.NEW}}
        )
    ) is not None:
        assert t2.run() == 1, "new doc run"
        assert (db_doc := col.find_one({"_id": doc["_id"]})) is not None, "doc found"
        assert (
            db_doc["code"] == DocumentCode.OK
        ), "new doc finished"

    assert t2.run() == 0, "no docs to run; dependent still pending"

    # blast AbsWorker one minute into the future, past the retry_after time
    ptime = mocker.patch("ampel.abstract.AbsWorker.time", return_value=time() + 60)

    assert t2.run() == 1
    assert ptime.called
    assert (db_doc := col.find_one({"_id": dependent_doc["_id"]})) is not None, "dependent doc found"
    assert db_doc["code"] == DocumentCode.OK


def test_queue_worker(
    mock_context: DevAmpelContext, mocker: MockerFixture, ampel_logger
):
    """
    Simulate a race conditions between parallel T2 workers
    """
    items: list[AbsProducer.Item] = []

    @mock_context.register_unit
    class DummyProducer(AbsProducer):
        def produce(self, item: AbsProducer.Item, delivery_callback=None):
            items.append(item)

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    @mock_context.register_unit
    class DummyConsumer(AbsConsumer):
        def consume(self) -> None | QueueItem:
            if not items:
                return None
            item = items.pop()
            return {"stock": item.stock, "t0": item.t0, "t1": item.t1, "t2": item.t2}

        def acknowledge(self, docs: Iterable[QueueItem]) -> None:
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    ack = mocker.patch.object(DummyConsumer, "acknowledge")

    mock_context.register_unit(QueueIngester)

    handler = make_tied_ingestion_handler(
        mock_context,
        ampel_logger,
        "DummyStateT2Unit",
        UnitModel(
            unit="QueueIngester",
            config={"producer": {"unit": "DummyProducer"}},
        ),
    )

    datapoints: list[dict[str, Any]] = [
        {"id": i, "stock": "stockystock", "body": {"thing": i + 1}} for i in range(3)
    ]

    with handler.ingester.group():
        handler.ingest(datapoints, [(0, True)], stock_id="stockystock", jm_extra={"alert": 123})

    t2 = T2QueueWorker(
        context=mock_context,
        consumer={"unit": "DummyConsumer"},
        raise_exc=True,
        process_name="t2",
        run_dependent_t2s=True,
        backoff_on_retry=[{"jitter": False, "factor": 10}],
    )

    assert t2.run() == 1

    assert ack.call_count == 1
    acks_iter = ack.call_args[0][0]
    assert isinstance(acks_iter, Iterable)
    acks = list(acks_iter)
    assert acks
    assert acks[0]["stock"][0]["stock"] == "stockystock"

    assert mock_context.db.get_collection("stock").count_documents({}) == 1
    assert mock_context.db.get_collection("t0").count_documents({}) == 3
    assert mock_context.db.get_collection("t1").count_documents({}) == 1
    t1_doc = mock_context.db.get_collection("t1").find_one({})
    assert t1_doc is not None
    assert len(t1_doc["meta"]) == 1

    docs = list(mock_context.db.get_collection("t2").find({"code": DocumentCode.OK}))
    assert len(docs) == 2
    assert len(docs[0]["body"]) == 1
    assert len(docs[0]["meta"]) == 2

    with handler.ingester.group():
        handler.ingest(datapoints, [(0, True)], stock_id="stockystock", jm_extra={"alert": 456})
    
    assert t2.run() == 1

    assert mock_context.db.get_collection("t1").count_documents({}) == 1
    t1_doc = mock_context.db.get_collection("t1").find_one({})
    assert t1_doc is not None
    assert len(t1_doc["meta"]) == 2, "meta entry added to t1 doc"
    docs = list(mock_context.db.get_collection("t2").find({"code": DocumentCode.OK}))
    assert len(docs) == 2
    assert len(docs[0]["body"]) == 1, "doc was not re-run"
    assert len(docs[0]["meta"]) == 3, "meta entry added to t2 doc"


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/AmpelLogger.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.09.2018
# Last Modified Date:  11.11.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys, logging, traceback
from os.path import basename
from sys import _getframe
from typing import TYPE_CHECKING, Any
from typing_extensions import Self

from ampel.types import ChannelId
from ampel.abstract.AbsContextManager import AbsContextManager
from ampel.log.handlers.AmpelStreamHandler import AmpelStreamHandler
from ampel.log.LightLogRecord import LightLogRecord
from ampel.log.LogFlag import LogFlag
from ampel.protocol.LoggingHandlerProtocol import AggregatingLoggingHandlerProtocol, LoggingHandlerProtocol
from ampel.util.hash import build_unsafe_dict_id


if TYPE_CHECKING:
	from ampel.mongo.update.var.DBLoggingHandler import DBLoggingHandler

ERROR = LogFlag.ERROR
WARNING = LogFlag.WARNING
SHOUT = LogFlag.SHOUT
INFO = LogFlag.INFO
VERBOSE = LogFlag.VERBOSE
DEBUG = LogFlag.DEBUG

if TYPE_CHECKING:
	from ampel.core.AmpelContext import AmpelContext

class AmpelLogger(AbsContextManager):

	loggers: dict[int | str, Self] = {}
	_counter: int = 0
	verbose: int = 0


	@classmethod
	def get_logger(cls, force_refresh: bool = False, **kwargs) -> Self:
		"""
		Get or create an :class:`AmpelLogger <ampel.log.AmpelLogger>` instance
		based on the provided keyword arguments.

		The logger is stored in the class-level ``loggers`` registry under a key:
		- If no arguments are provided, the registry key defaults to ``"default"``.
		- If ``name`` is provided in ``kwargs``, it serves as registry key.
		- Otherwise, a unique identifier is derived from the keyword arguments
		  and used as the key.

		Once the key is determined:
		- If a logger with the same key already exists, that instance is returned.
		- If no logger exists for the key, a new one is created and stored.
		- If ``force_refresh`` is ``True``, a new logger is always created and
		  replaces any existing one for the key.

		Important notes:
		- This method is intended primarily for cases where one wants to reuse
		  a logger across calls. It is not suitable if you need to pass
		  instantiated handler objects to the logger. In such cases, call the
		  :class:`AmpelLogger` constructor directly.
		- All values in ``kwargs`` must be serializable. Passing non-serializable
		  objects will break the identifier generation.

		:param force_refresh: If ``True``, always create a new logger instance.
		:param kwargs: Additional keyword arguments passed to the constructor.
		:returns: An :class:`AmpelLogger` instance.

		Example usage::

			# Default logger with "default" key
			logger = AmpelLogger.get_logger()

			# Logger with explicit console level
			logger = AmpelLogger.get_logger(console={'level': DEBUG})

			# Force creation and caching of a new logger
			logger = AmpelLogger.get_logger(force_refresh=True, console={'level': INFO})

			# Logger with explicit name
			logger = AmpelLogger.get_logger(name="custom_logger")
		"""

		if not kwargs:
			kwargs["name"] = "default"

		if "name" not in kwargs:
			kwargs["name"] = build_unsafe_dict_id(kwargs)

		if force_refresh or kwargs["name"] not in cls.loggers:
			cls.loggers[kwargs["name"]] = cls(**kwargs)

		return cls.loggers[kwargs["name"]]


	@classmethod
	def has_logger(cls, name: str) -> bool:
		return name in cls.loggers


	@classmethod
	def from_profile(cls,
		context: 'AmpelContext', profile: str, run_id: int | None = None, **kwargs
	) -> Self:
		"""
		Creates and returns an AmpelLogger instance configured from a named logging profile.

		If no logger exists with the same combination of profile name, run ID, context UUID,
		and additional keyword arguments, a new logger is created and configured.
		Otherwise, a previously created logger with matching parameters
		is returned from the internal cache.

		The logger is configured based on the specified profile from the ampel configuration.
		Depending on the profile, it may include handlers such as DBLoggingHandler (requires run_id)
		and AmpelStreamHandler for console output.

		Parameters:
		- context (AmpelContext): Provides access to configuration and database connections.
		- profile (str): Name of the logging profile to load from config.
		- run_id (int | None): Required if the profile includes a DB logging handler.
		- **kwargs: Additional keyword arguments passed to the logger constructor.

		Returns:
		- AmpelLogger: A logger instance configured according to the specified profile.

		Raises:
		- ValueError: If the profile requires a DBLoggingHandler and no run_id is provided.
		"""

		logger_id = f"{profile}_{run_id}_{build_unsafe_dict_id(kwargs)}_{context.uuid}"
		kwargs["name"] = logger_id

		if logger_id in cls.loggers:
			return cls.get_logger(**kwargs)

		# Note: "console": False ensures DBLoggingHandler is inserted first
		logger = cls.get_logger(**(kwargs | {"console": False}))

		profile_dict = context.config.get(f'logging.{profile}', dict, raise_exc=True)
		if "db" in profile_dict:

			# avoid circular import
			from ampel.mongo.update.var.DBLoggingHandler import DBLoggingHandler # noqa: PLC0415

			if run_id is None:
				raise ValueError("Parameter 'run_id' is required when log_profile requires db logging handler")

			logger.addHandler(
				DBLoggingHandler(context.db, run_id, **profile_dict['db'])
			)

		if "console" in profile_dict:
			logger.addHandler(
				AmpelStreamHandler(**profile_dict['console'])
			)

		return logger


	@staticmethod
	def get_console_level(context: 'AmpelContext', profile: str) -> int | None:

		handlers = context.config.get(f'logging.{profile}', dict, raise_exc=True)

		if "console" in handlers:
			if 'level' in handlers['console']:
				return handlers['console']['level']
			return LogFlag.INFO.__int__()

		return None


	@classmethod
	def has_verbose_console(cls, context: 'AmpelContext', profile: str) -> bool:

		if lvl := cls.get_console_level(context, profile):
			return lvl < INFO
		return False


	def __init__(self,
		name: int | str = 0,
		base_flag: LogFlag | None = None,
		handlers: list[LoggingHandlerProtocol | AggregatingLoggingHandlerProtocol] | None = None,
		channel: ChannelId | list[ChannelId] | None = None,
		# See AmpelStreamHandler annotations for more details
		console: bool | dict[str, Any] | None = True
	) -> None:

		self.name = name
		self.base_flag = base_flag.__int__() if base_flag else 0
		self.handlers = handlers or []
		self.channel = channel
		self.level = 0
		self.fname = _getframe().f_code.co_filename

		if console:
			shargs: dict[str, Any] = {"color": False} if (base_flag and LogFlag.UNIT in base_flag) else {}
			if isinstance(console, dict):
				shargs |= console
			self.addHandler(AmpelStreamHandler(**shargs))
		else:
			self.provenance = False

		self._auto_level()


	def __exit__(self, exc_type, exc_value, traceback) -> None:
		self.flush()


	def _auto_level(self) -> None:

		self.level = min([h.level for h in self.handlers]) if self.handlers else 0
		if self.level < INFO:
			self.verbose = 2 if self.level < VERBOSE else 1
		elif self.verbose != 0:
			self.verbose = 0


	def addHandler(self, handler: LoggingHandlerProtocol) -> None:

		self.level = min(handler.level, self.level)

		if isinstance(handler, AmpelStreamHandler) and handler.provenance:
			self.provenance = True

		if self.level < INFO:
			self.verbose = 2 if self.level < VERBOSE else 1

		self.handlers.append(handler)


	def removeHandler(self, handler: LoggingHandlerProtocol) -> None:
		self.handlers.remove(handler)
		self._auto_level()


	def get_db_logging_handler(self) -> 'DBLoggingHandler | None':
		# avoid circular import
		from ampel.mongo.update.var.DBLoggingHandler import (  # noqa: PLC0415
			DBLoggingHandler,
		)
		for el in self.handlers:
			if isinstance(el, DBLoggingHandler):
				return el
		return None


	def break_aggregation(self) -> None:
		for el in self.handlers:
			if isinstance(el, AggregatingLoggingHandlerProtocol | AmpelStreamHandler):
				el.break_aggregation()


	def error(self, msg: str | dict[str, Any], *args,
		exc_info: Exception | None = None,
		stack_info: bool = False,
		stacklevel: int = 1,
		extra: dict[str, Any] | None = None,
	):
		self.log(ERROR, msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extra=extra)


	def warn(self, msg: str | dict[str, Any], *args,
		stack_info: bool = False,
		stacklevel: int = 1,
		extra: dict[str, Any] | None = None,
	):
		if self.level <= WARNING:
			self.log(WARNING, msg, *args, stack_info=stack_info, stacklevel=stacklevel, extra=extra)


	def info(self, msg: str | dict[str, Any] | None, *args,
		stack_info: bool = False,
		stacklevel: int = 1,
		extra: dict[str, Any] | None = None,
	) -> None:
		if self.level <= INFO:
			self.log(INFO, msg, *args, stack_info=stack_info, stacklevel=stacklevel, extra=extra)


	def debug(self, msg: str | dict[str, Any] | None, *args,
		stack_info: bool = False,
		stacklevel: int = 1,
		extra: None | dict[str, Any] = None,
	):
		if self.level <= DEBUG:
			self.log(DEBUG, msg, *args, stack_info=stack_info, stacklevel=stacklevel, extra=extra)


	def handle(self, record: LightLogRecord | logging.LogRecord) -> None:
		for h in self.handlers:
			if record.levelno >= h.level:
				h.handle(record)


	def flush(self) -> None:
		for h in self.handlers:
			h.flush()


	def log(self,
		lvl: int, msg: str | dict[str, Any] | None, *args,
		exc_info: bool | Exception | None = None,
		stack_info: bool = False,
		stacklevel: int = 1,
		extra: dict[str, Any] | None = None,
	):

		if args and isinstance(msg, str):
			msg = msg % args

		record = LightLogRecord(name=self.name, levelno=lvl | self.base_flag, msg=msg)

		if lvl > WARNING or self.provenance:
			frame = _getframe(stacklevel) # logger.log(...) was called directly
			if frame.f_code.co_filename == self.fname:
				frame = _getframe(stacklevel+1) # logger.info(...), logger.debug(...) was used
			record.__dict__['filename'] = basename(frame.f_code.co_filename)
			record.__dict__['lineno'] = frame.f_lineno

		if extra:
			extra = dict(extra)
			if (stock := extra.pop("stock", None)):
				record.stock = stock
			if (channel := (extra.pop("channel", None) or self.channel)):
				record.channel = channel
			if (unit := (extra.pop("unit", None))):
				record.unit = unit
			record.extra = extra

		if exc_info:

			if exc_info == 1:
				exc_info = sys.exc_info() # type: ignore[assignment]
				lines = traceback.format_exception(*sys.exc_info())
			elif isinstance(exc_info, tuple):
				lines = traceback.format_exception(*sys.exc_info())
			elif isinstance(exc_info, Exception):
				lines = traceback.format_exception(
					type(exc_info), exc_info, exc_info.__traceback__
				)
			else:
				lines = []

			erec = AmpelLogger.fork_rec(record, "\n")
			for h in self.handlers:
				h.handle(erec)

			for el in lines:
				for l in el.split('\n'):
					if not l:
						continue
					erec = AmpelLogger.fork_rec(record, l)
					for h in self.handlers:
						h.handle(erec)

			if record.msg:
				rec2 = AmpelLogger.fork_rec(record, "-" * len(record.msg))
				for h in self.handlers:
					h.handle(record)
					h.handle(rec2)

			return

		for h in self.handlers:
			if record.levelno >= h.level:
				h.handle(record)


	@staticmethod
	def fork_rec(orig: LightLogRecord, msg: str) -> LightLogRecord:
		new_rec = LightLogRecord(name=0, msg=None, levelno=0)
		for k, v in orig.__dict__.items():
			new_rec.__dict__[k] = v
		new_rec.msg = msg
		return new_rec



from collections.abc import Callable, Generator, Iterable
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Literal

from ampel.abstract.AbsContextManager import AbsContextManager
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.base.decorator import abstractmethod
from ampel.content.DataPoint import DataPoint
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.core.ContextUnit import ContextUnit
from ampel.log.AmpelLogger import AmpelLogger
from ampel.protocol.StockIngesterProtocol import StockIngesterProtocol
from ampel.types import Traceless

if TYPE_CHECKING:
	pass

class AbsIngester(AbsContextManager, ContextUnit, abstract=True):
	error_callback: Traceless[None | Callable[[], None]] = None
	acknowledge_callback: Traceless[None | Callable[[Iterable[Any]], None]] = None

	raise_exc: bool = False

	run_id: Traceless[int]
	tier: Traceless[Literal[-1, 0, 1, 2, 3]]
	process_name: Traceless[str]
	logger: Traceless[AmpelLogger]

	@contextmanager
	@abstractmethod
	def group(self, acknowledge_messages: None | Iterable[Any] = None) -> Generator:
		"""
		Ensure that documents ingested in this context are grouped together

		:param acknowledge_messages: messages to be passed to
			acknowledge_callback when documents ingested in the context are
			delivered
		"""
		raise NotImplementedError

	@property
	@abstractmethod
	def stock(self) -> StockIngesterProtocol:
		raise NotImplementedError

	@property
	@abstractmethod
	def t0(self) -> AbsDocIngester[DataPoint]:
		raise NotImplementedError

	@property
	@abstractmethod
	def t1(self) -> AbsDocIngester[T1Document]:
		raise NotImplementedError

	@property
	@abstractmethod
	def t2(self) -> AbsDocIngester[T2Document]:
		raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t2/T2QueueWorker.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                unspecified
# Last Modified Date:  unspecified
# Last Modified By:    jvs

import gc
from collections.abc import Generator, Sequence
from time import time
from typing import Any, Literal, TypedDict, overload

from bson import ObjectId
from mongomock.filtering import filter_applies

from ampel.abstract.AbsIngester import AbsIngester
from ampel.abstract.AbsWorker import stat_time
from ampel.content.DataPoint import DataPoint
from ampel.content.MetaRecord import MetaRecord
from ampel.content.StockDocument import StockDocument
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.core.EventHandler import EventHandler
from ampel.enum.DocumentCode import DocumentCode
from ampel.log import AmpelLogger, LogFlag
from ampel.model.UnitModel import UnitModel
from ampel.queue.AbsConsumer import AbsConsumer
from ampel.t2.T2Worker import T2Worker
from ampel.types import (
	DataPointId,
	OneOrMany,
	StockId,
	T2Link,
	Tag,
	UBson,
)


class QueueItem(TypedDict):
	stock: Sequence[StockDocument]
	t0: Sequence[DataPoint]
	t1: Sequence[T1Document]
	t2: Sequence[T2Document]



class T2QueueWorker(T2Worker):

	consumer: UnitModel
	ingester: UnitModel = UnitModel(unit="MongoIngester")
	
	# Must run dependent t2s, as no other worker will see them
	run_dependent_t2s: Literal[True] = True
	pre_check: Literal[False] = False

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)

		self._current_item: None | QueueItem = None
	
	def proceed(self, event_hdlr: EventHandler) -> int:
		""" :returns: number of t2 docs processed """

		event_hdlr.set_tier(self.tier)
		run_id = event_hdlr.get_run_id()

		if self.send_beacon:
			self.col_beacon.update_one(
				{'_id': self.beacon_id},
				{'$set': {'timestamp': int(time())}}
			)

		# Loop variables
		doc_counter = 0
		garbage_collect = self.garbage_collect
		doc_limit = self.doc_limit

		with (
			AmpelLogger.from_profile(
				self.context, self.log_profile, run_id,
				base_flag = getattr(LogFlag, f'T{self.tier}') | LogFlag.CORE | self.base_log_flag
			) as logger,
			self._run_until_signal(run_id, logger) as stop_token,
			self.context.loader.new(
				self.consumer,
				stop=stop_token,
				unit_type=AbsConsumer,
			) as consumer,
			self.context.loader.new_context_unit(
				self.ingester,
				context = self.context,
				run_id = run_id,
				tier = self.tier,
				process_name = self.process_name,
				error_callback = stop_token.set,
				acknowledge_callback = consumer.acknowledge,
				logger = logger,
				sub_type = AbsIngester,
			) as ingester
		):

			# Process docs until next() returns None (breaks condition below)
			while not stop_token.is_set():

				# get t1/t2 document (code is usually NEW or NEW_PRIO), excluding
				# docs with retry times in the future
				with stat_time.labels(self.tier, "consume", None).time():
					item: None | QueueItem = consumer.consume()

				# No match
				if item is None:
					if not stop_token.is_set():
						logger.log(LogFlag.SHOUT, "No more docs to process")
					break

				self._current_item = item

				input_docs = item["t2"]
				# Replace inputs with resolved version from the database.
				# Doing this here allows process_doc() to find
				# already-processed docs if it recurse into
				# load_input_docs() for tied units.
				item["t2"] = [self._sub_existing_doc(doc) for doc in item["t2"]]

				with ingester.group([item]):
					for input_doc, doc in zip(input_docs, item["t2"], strict=True):
						if doc is input_doc:
							# not found in the database; process for the first time
							with stat_time.labels(self.tier, "process_doc", doc["unit"]).time():
								self.process_doc(doc, ingester, logger)
					doc_counter += 1

					for stock in item["stock"]:
						ingester.stock.ingest(stock)
					for dp in item["t0"]:
						ingester.t0.ingest(dp)
					for t1 in item["t1"]:
						ingester.t1.ingest(t1)
					# NB: ingest the input docs as they were before
					# substitution in order to pick up any requested updates
					# to meta, channels, tags, expiry, etc.
					for t2 in input_docs:
						ingester.t2.ingest(t2)

				# Check possibly defined doc_limit
				if doc_limit and doc_counter >= doc_limit:
					break

				if garbage_collect:
					gc.collect()

		event_hdlr.add_extra(docs=doc_counter)
		return doc_counter

	def _sub_existing_doc(self, doc: T2Document) -> T2Document:
		"""replace doc with a resolved copy from the database if it exists"""
		match = {
			'code': DocumentCode.OK,
			'stock': doc['stock'],
			'unit': doc['unit'],
			'config': doc['config'],
			'link': doc['link']
		}

		if 'origin' in doc:
			match['origin'] = doc['origin']

		db_doc: None | T2Document
		if db_doc := self.col.find_one(match):
			# merge the existing doc with the new one for consistency
			db_doc["meta"] = [*db_doc["meta"], *doc.get("meta", [])]
			db_doc["body"] = [*db_doc["body"], *doc.get("body", [])]
			db_doc["tag"] = list(set(db_doc.get("tag", [])).union(doc.get("tag", [])))
			db_doc["channel"] = list(set(db_doc.get("channel", [])).union(doc.get("channel", [])))

			return db_doc
		return doc

	def update_doc(self,
		doc: T2Document,
		meta: MetaRecord,
		logger: AmpelLogger, *,
		body: UBson = None,
		tag: None | OneOrMany[Tag] = None,
		code: int = 0
	) -> None:
		# update doc in place; changes will be propagated by the ingester as a block
		doc["code"] = code
		doc["meta"] = [*doc["meta"], meta]
		if body:
			doc["body"] = [*doc["body"], body]
		if tag:
			doc["tag"] = list(
				set([tag] if isinstance(tag, Tag) else tag)
				.union(doc.get("tag", []))
			)

	def load_stock(self, stock: StockId) -> None | StockDocument:
		"""Load stock document from current message"""
		return (
			next((d for d in self._current_item["stock"] if d["stock"] == stock), None)
			if self._current_item is not None else None
		) or super().load_stock(stock)

	@overload
	def load_t0(self, stock: StockId | Sequence[StockId], t1_dps_ids: DataPointId) -> None | DataPoint: ...

	@overload
	def load_t0(self, stock: StockId | Sequence[StockId], t1_dps_ids: Sequence[DataPointId]) -> list[DataPoint]: ...

	def load_t0(self, stock: StockId | Sequence[StockId], t1_dps_ids: DataPointId | Sequence[DataPointId]) -> None | DataPoint | list[DataPoint]:
		"""Load datapoints from database"""
		if isinstance(t1_dps_ids, DataPointId):
			return (
				next((d for d in self._current_item["t0"] if d["stock"] == stock and d["id"] == t1_dps_ids), None)
				if self._current_item is not None else None
			) or super().load_t0(stock, t1_dps_ids)
		targets = set(t1_dps_ids)
		datapoints = (
			{dp["id"]: dp for dp in self._current_item["t0"] if dp["stock"] == stock and dp["id"] in targets}
			if self._current_item is not None else {}
		)
		# fall back to database if some datapoints are missing
		if missing := targets.difference(datapoints.keys()):
			return list(datapoints.values()) + super().load_t0(stock, list(missing))
		return list(datapoints.values())

	def load_t1(self, stock: StockId | Sequence[StockId], link: T2Link) -> None | T1Document:
		"""Load T1 document from database"""
		return (
			next((d for d in self._current_item["t1"] if d["stock"] == stock and d["link"] == link), None)
			if self._current_item is not None else None
		) or super().load_t1(stock, link)

	def load_t2(self, query: dict[str, Any], for_update: bool=False) -> Generator[T2Document]:
		"""
		Load T2 documents from database
								  |  for_update  |
		condition                 | true | false |
		--------------------------+------+-------+
		not in db                 | item | item  |
		in db, code OK            | -    | db    |
		in db, code in code_match | item | item  |
		in db, other code         | item | item  |
		"""

		if self._current_item is not None:
			if for_update:
				# want to run dependent t2, but already resolved
				if self.col.count_documents({"code": DocumentCode.OK} | query):
					return
				for doc in self._current_item["t2"]:
					if doc["code"] != DocumentCode.OK and filter_applies(query, doc):
						# prevent this doc from being returned by a call with for_update=False
						if "_id" not in doc:
							doc["_id"] = ObjectId()  # type: ignore[typeddict-unknown-key]
						yield doc
			else:
				count = 0
				for count, doc in enumerate(super().load_t2(query, for_update), 1):  # noqa: B007
					yield doc
				if count == 0:
					# return t2 docs that are not in the database
					for doc in self._current_item["t2"]:
						if filter_applies(query, doc):
							yield doc
		else:
			raise RuntimeError("load_t2 called outside of consume loop")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/pretty.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                Unspecified
# Last Modified Date:  06.09.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import html
import math
import re
import sys
from contextlib import contextmanager
from datetime import timedelta
from math import isinf
from time import time

from ampel.protocol.LoggerProtocol import LoggerProtocol


# copied from https://stackoverflow.com/a/56497521/104668
def prettyjson(obj, indent=2, maxlinelength=80):
	"""Renders JSON content with indentation and line splits/concatenations to fit maxlinelength.
	Only dicts, lists and basic types are supported"""
	items, _ = getsubitems(obj, itemkey="", islast=True, maxlinelength=maxlinelength, level=0)
	return indentitems(items, indent, level=0)


def getsubitems(obj, itemkey, islast, maxlinelength, level):

	items = []
	is_inline = True # at first, assume we can concatenate the inner tokens into one line

	isdict = isinstance(obj, dict)
	islist = isinstance(obj, list)
	istuple = isinstance(obj, tuple)
	isbasictype = not (isdict or islist or istuple)

	# build json content as a list of strings or child lists
	if isbasictype:
		# render basic type
		keyseparator = "" if itemkey == "" else ": "
		itemseparator = "" if islast else ","
		items.append(itemkey + keyseparator + basictype2str(obj) + itemseparator)

	else:
		# render lists/dicts/tuples
		if isdict:
			opening, closing, keys = ("{", "}", iter(obj.keys()))
		elif islist:
			opening, closing, keys = ("[", "]", range(0, len(obj)))
		elif istuple:
			opening, closing, keys = ("[", "]", range(0, len(obj)))	# tuples are converted into json arrays

		if itemkey != "":
			opening = itemkey + ": " + opening
		if not islast:
			closing += ","

		itemkey = ""
		subitems = []

		# get the list of inner tokens
		for (i, k) in enumerate(keys):
			islast_ = i == len(obj) - 1
			itemkey_ = ""
			if isdict:
				itemkey_ = basictype2str(k)
			inner, is_inner_inline = getsubitems(obj[k], itemkey_, islast_, maxlinelength, level + 1)
			subitems.extend(inner)						# inner can be a string or a list
			is_inline = is_inline and is_inner_inline	 # if a child couldn't be rendered inline, then we are not able either

		# fit inner tokens into one or multiple lines, each no longer than maxlinelength
		if is_inline:
			multiline = True

			# in Multi-line mode items of a list/dict/tuple can be rendered in multiple lines if they don't fit on one.
			# suitable for large lists holding data that's not manually editable.

			# in Single-line mode items are rendered inline if all fit in one line, otherwise each is rendered in a separate line.
			# suitable for smaller lists or dicts where manual editing of individual items is preferred.

			# this logic may need to be customized based on visualization requirements:
			if (isdict):
				multiline = False
			if (islist):
				multiline = True

			if (multiline):
				lines = []
				current_line = ""

				for (i, item) in enumerate(subitems):
					item_text = item
					if i < len(inner) - 1:
						item_text = item + ","

					if len(current_line) > 0:
						try_inline = current_line + " " + item_text
					else:
						try_inline = item_text

					if (len(try_inline) > maxlinelength):
						# push the current line to the list if maxlinelength is reached
						if len(current_line) > 0:
							lines.append(current_line)
						current_line = item_text
					else:
						# keep fitting all to one line if still below maxlinelength
						current_line = try_inline

					# Push the remainder of the content if end of list is reached
					if (i == len(subitems) - 1):
						lines.append(current_line)

				subitems = lines
				if len(subitems) > 1:
					is_inline = False
			else: # single-line mode
				totallength = len(subitems) - 1   # spaces between items
				for item in subitems:
					totallength += len(item)
				if (totallength <= maxlinelength):
					str = ""
					for item in subitems:
						str += item + " "  # insert space between items, comma is already there
					subitems = [str.strip()]			   # wrap concatenated content in a new list
				else:
					is_inline = False


		# attempt to render the outer brackets + inner tokens in one line
		if is_inline:
			item_text = ""
			if len(subitems) > 0:
				item_text = subitems[0]
			if len(opening) + len(item_text) + len(closing) <= maxlinelength:
				items.append(opening + item_text + closing)
			else:
				is_inline = False

		# if inner tokens are rendered in multiple lines already, then the outer brackets remain in separate lines
		if not is_inline:
			items.append(opening)	   # opening brackets
			items.append(subitems)	  # Append children to parent list as a nested list
			items.append(closing)	   # closing brackets

	return items, is_inline


def basictype2str(obj):
	if (
		isinstance(obj, str) or
		obj.__class__.__name__ == "ObjectId" or
		(isinstance(obj, float) and isinf(obj))
	):
		return "\"" + str(obj) + "\""
	if isinstance(obj, bool):
		return "true" if obj else "false"
	return str(obj)


def indentitems(items, indent, level):
	"""Recursively traverses the list of json lines, adds indentation based on the current depth"""
	res = ""
	indentstr = " " * (indent * level)
	for (i, item) in enumerate(items):
		if isinstance(item, list):
			res += indentitems(item, indent, level + 1)
		else:
			islast = (i == len(items) - 1)
			# no new line character after the last rendered line
			if level == 0 and islast:
				res += indentstr + item
			else:
				res += indentstr + item + "\n"
	return res

# End of prettyjson
###################

# Notebook goodies
def set_bold(s: str, match: str):

	from IPython.display import HTML  # noqa: PLC0415
	out = []

	for el in s.split("\n"):
		if re.match(match, el):
			out.append(f"<b>{html.escape(el)}</b>")
		else:
			out.append(html.escape(el))
	return HTML("<pre style='font-size: 13px'>" + "<br/>".join(out) + "</pre>")


@contextmanager
def out_stack():
	"""
	with out_stack():
		raise ValueError("Clean and concise")
	"""
	default_value = getattr(sys, "tracebacklimit", 1000)
	sys.tracebacklimit = 0
	print(" ")  # noqa: T201
	yield
	sys.tracebacklimit = default_value


def human_format(num, precision=2, suffixes=('', 'K', 'M', 'G', 'T', 'P')):
	"""
	In []: human_format(1000000, precision=0)
	Out[]: '1M'

	In []: human_format(30000, precision=0)
	Out[]: '30K'

	In []: human_format(123456)
	Out[]: '123.46K'
	"""
	m = int(math.log10(num) // 3)
	return f'{num/1000.0**m:.{precision}f}{suffixes[m]}'


def get_time_delta(start: float, cut: None | int = -7) -> str:
	return str(timedelta(seconds=time()-start))[:-7]


class TimeFeedback:
	"""
	Usage:
	with TimeFeedback("2d binning"):
		do_it()
	"""

	def __init__(self,
		feedback: str = "Time elapsed",
		cut: None | int = -7,
		logger: None | LoggerProtocol = None,
		enter_feedback: bool = False
	):
		self.feedback = feedback
		self.logger = logger
		self.cut = cut
		self.enter_feedback = enter_feedback

	def __enter__(self):
		self.start = time()
		if self.enter_feedback:
			self._print(
				f"Measuring time for '{self.feedback}'"
				if self.feedback != "Time elapsed"
				else "Starting time measurement"
			)
		return self

	def __exit__(self, *args):
		self._print(self.feedback + ": " + get_time_delta(self.start, self.cut))

	def _print(self, s: str) -> None:
		if self.logger:
			self.logger.info(s)
		print(s)  # noqa: T201
